package GitLab::Users;

use utf8;
use strict;
use warnings;
use vars qw($VERSION);

use GitLab::API;
use Log::Any        qw($log);

our $VERSION = v10.4.1;

my $requests = {
    users => {
        method      => "GET",
        path        => "/users",
        query       => [qw(
            username
            search
            blocked
            active
            extern_uid
            provider
            created_before
            created_after
            custom_attributes
            external
        )],
        paginated   => 1,
    },

    user_by_id => {
        method      => "GET",
        path        => "/users/<uid>",
    },

    user_create     => {
        method      => "POST",
        path        => "/users",
        required    => [qw(
            email
            username
            name
        )],
        optional    => [qw(
            password
            reset_password
            skype
            linkedin
            twitter
            website_url
            organization
            projects_limit
            extern_uid
            provider
            bio
            location
            admin
            can_create_group
            skip_confirmation
            external
            avatar
        )],
    },

    user_update     => {
        method      => "PUT",
        path        => "/users/<uid>",
        # update can cause Conflict, but API always returns 404
        ret         => {
            404 => "Not found or Conflict"
        },
        optional    => [qw(
            email
            username
            name
            password
            skype
            linkedin
            twitter
            website_url
            organization
            projects_limit
            extern_uid
            provider
            bio
            location
            admin
            can_create_group
            skip_confirmation
            external
            avatar
        )],
    },

    user_delete     => {
        method      => "DELETE",
        path        => "/users/<uid>",
    },

    # SSH keys
    self_ssh_keys => {
        method      => "GET",
        path        => "/user/keys",
        paginated   => 1,
    },

    user_ssh_keys => {
        method      => "GET",
        path        => "/users/<uid>/keys",
        paginated   => 1,
    },

    self_get_ssh_key => {
        method      => "GET",
        path        => "/user/keys/<keyid>",
    },

    self_add_ssh_key => {
        method      => "POST",
        path        => "/user/keys",
        required    => [qw(
            title
            key
        )],
    },

    user_add_ssh_key => {
        method      => "POST",
        path        => "/users/<uid>/keys",
        required    => [qw(
            title
            key
        )],
    },

    self_delete_ssh_key => {
        method      => "DELETE",
        path        => "/user/keys/<keyid>",
    },

    user_delete_ssh_key => {
        method      => "DELETE",
        path        => "/users/<uid>/keys/<keyid>",
    },

    # GPG keys
    self_gpg_keys => {
        method      => "GET",
        path        => "/user/gpg_keys",
        paginated   => 1,
    },

    user_gpg_keys   => {
        method      => "GET",
        path        => "/users/<uid>/gpg_keys",
        paginated   => 1,
    },

    self_get_gpg_key => {
        method      => "GET",
        path        => "/user/gpg_keys/<keyid>",
    },

    user_get_gpg_key => {
        method      => "GET",
        path        => "/users/<uid>/gpg_keys/<keyid>",
    },

    self_add_gpg_key => {
        method      => "POST",
        path        => "/user/gpg_keys",
        required    => [qw(
            key
        )],
    },

    user_add_gpg_key => {
        method      => "POST",
        path        => "/users/<uid>/gpg_keys",
        required    => [qw(
            key
        )],
    },

    self_delete_gpg_key => {
        method      => "DELETE",
        path        => "/user/gpg_keys/<keyid>",
    },

    user_delete_gpg_key => {
        method      => "DELETE",
        path        => "/users/<uid>/gpg_keys/<keyid>",
    },

    # E-mails
    self_get_emails => {
        method      => "GET",
        path        => "/user/emails",
        paginated   => 1,
    },

    user_get_emails => {
        method      => "GET",
        path        => "/users/<uid>/emails",
        paginated   => 1,
    },

    self_get_email => {
        method      => "GET",
        path        => "/user/emails/<eid>",
    },

    self_add_email => {
        method      => "POST",
        path        => "/user/emails",
        required    => [ qw!email! ],
    },

    user_add_email => {
        method      => "POST",
        path        => "/users/<uid>/emails",
        required    => [ qw!email! ],
    },

    self_delete_email => {
        method      => "DELETE",
        path        => "/user/emails/<eid>",
    },

    user_delete_email => {
        method      => "DELETE",
        path        => "/users/<uid>/emails/<eid>",
    },

    # Tokens
    user_tokens => {
        method      => "GET",
        path        => "/users/<uid>/impersonation_tokens",
        query       => [qw(
            state
        )],
        paginated   => 1,
    },

    user_get_token => {
        method      => "GET",
        path        => "/users/<uid>/impersonation_tokens/<tokenid>",
    },

    user_add_token => {
        method      => "POST",
        path        => "/users/<uid>/impersonation_tokens",
        required    => [qw(
            name
        )],
        query       => [qw(
            expires_at
            scopes
        )],
    },

    user_delete_token => {
        method      => "DELETE",
        path        => "/users/<uid>/impersonation_tokens/<tokenid>",
    },

    # Block/unblock
    user_block      => {
        method      => "POST",
        path        => "/users/<uid>/block",
    },

    user_unblock    => {
        method      => "POST",
        path        => "/users/<uid>/unblock",
    },

    # User activities
    user_activities => {
        method      => "GET",
        path        => "/user/activities",
        paginated   => 1,
    },
};

sub import {
    $log->debug("initializing " . __PACKAGE__);
    while (my ($name, $tmpl) = each(%$requests)) {
        $tmpl->{name} = $name unless exists $tmpl->{name};
        GitLab::API->register($tmpl);
    }
}

1;

__END__

=head1 NAME

GitLab::Users - implements user API calls

See L<GitLab API -- Users|http://doc.gitlab.com/ce/api/users.html> for details and
response formats.

=head1 VERSION

Implements API calls for GitLab CE C<v10.4.1>.

=head1 SYNOPSIS

    use GitLab::API;
    use GitLab::Users;
    # all calls from GitLab::Users are "injected" into the GitLab::API as methods

    my $api = GitLab::API->new();

    my $john = $api->users(username => "john_doe");

    $api->user_create(
        email       => "jane34@somewhere.com",
        password    => "s3cre1",
        username    => "jane.doe",
        name        => "Jane Doe"
    );

=head1 DESCRIPTION

=head2 Notation

All methods take a hash of arguments as specified in the documentation for
L<GitLab::API/exec_request>, but to simplify the documentation,
we will use the following notation:

    users( :username, :search )

which means that the C<users> method takes a hash with two keys, C<username>
and C<search>, so it can be called as

    $gitlab->users(search => "john_doe");

Optional arguments are denoted as C< [:username] >.
Note that not all optional arguments are listed.
Please refer to the official documentation for the full list.

=head2 User CRUD operations

=over

=item L<users()|https://docs.gitlab.com/ce/api/users.html#list-users>

    $users = $gitlab->users( [:username], [:search] );

This method returns all users. The optional arguments can be used
to filter out the results:

    username        search for a user by his username (exact match)
    search          search for users with username or email (substring match)

Note that the returned value is still an array even when C<< users(username => $username) >>
always returns at most one result.

=item L<user_by_id()|https://docs.gitlab.com/ce/api/users.html#single-user>

    $user = $gitlab->user_by_id( :uid );

Returns a user with the given C<uid>.

=item L<user_create()|https://docs.gitlab.com/ce/api/users.html#user-creation>

    $user = $gitlab->user_create( :email, :password, :username, :name );

Creates a new user and if succeeds, returns the generated account information.

    email               primary email, must not exist in GitLab,
    password            user password,
    username            login, must not exist in GitLab,
    name                display name

The function can take many optional arguments. Please see the API documentation for
the full list. Most interesting arguments are:

    admin               create an admin account (bool)
    can_create_group    the user will be able to create group (bool)
    external            create an external account (bool)
    projects_limit      maximum number of projects the user can own

=item L<user_update()|https://docs.gitlab.com/ce/api/users.html#user-modification>

    $user = $gitlab->user_update( :uid );

Updates the user with the given C<uid>. It takes the same arguments as C<user_create>
except they are all optional. Note that though it is possible to change user's primary e-mail
and username, it is probably not a very sane thing to do.

Also, from documentation:

    Note, at the moment this method does only return a 404 error, even in cases where
    a 409 (Conflict) would be more appropriate, e.g. when renaming the email address to some existing one.

=item L<user_delete()|https://docs.gitlab.com/ce/api/users.html#user-deletion>

    $gitlab->user_delete( :uid );

Deletes the user with the given C<uid>.

Note that

    calling this function for a non-existent user id still returns a status code 200 OK.
    The JSON response differs if the user was actually deleted or not.

(from the GitLab API documentation).

=item L<user()|https://docs.gitlab.com/ce/api/users.html#user>

B<This method was removed from the implementation of API>.
It still exists in the GitLab API, but was confusing to use.
Use L<GitLab::API::whoami()|GitLab::API/whoami> method instead.

=back

=head2 SSH keys

=over

=item L<self_ssh_keys()|https://docs.gitlab.com/ce/api/users.html#list-ssh-keys>

=item L<user_ssh_keys()|https://docs.gitlab.com/ce/api/users.html#single-ssh-key>

    $keys = $gitlab->self_ssh_keys();
    $keys = $gitlab->user_ssh_keys( :uid );

Returns a list of SSH keys for the authenticated user or the user with the
given C<uid>.

=item L<self_get_ssh_key()|https://docs.gitlab.com/ce/api/users.html#list-ssh-keys-for-user>

    $key = $gitlab->self_get_ssh_key( :keyid );

For the user that makes the request, the function returns the SSH key with the given C<keyid>.

B<Note>: there is no metod to get a single SSH key of another user.
Use combination of L<GitLab::API::sudo()|GitLab::API/sudo()> and
L<GitLab::Users::self_get_ssh_key()|GitLab::Users/self_get_ssh_key()>
instead.

=item L<self_add_ssh_key()|https://docs.gitlab.com/ce/api/users.html#add-ssh-key>

=item L<user_add_ssh_key()|https://docs.gitlab.com/ce/api/users.html#add-ssh-key-for-user>

    $key = $gitlab->self_add_ssh_key( :title, :key );
    $key = $gitlab->user_add_ssh_key( :uid, :title, :key );

Adds a new key for the user that makes the request or the given user.
Required arguments are

    title               the title of the key
    key                 public SSH key

=item L<self_delete_ssh_key()|https://docs.gitlab.com/ce/api/users.html#delete-ssh-key-for-current-user>

=item L<user_delete_ssh_key()|https://docs.gitlab.com/ce/api/users.html#delete-ssh-key-for-given-user>

    $gitlab->self_delete_ssh_key( :keyid );
    $gitlab->user_delete_ssh_key( :uid, :keyid );

Removes the SSH key with the given C<keyid> for the user that makes the request
or the given user.

=back

=head2 GPG keys

=over

=item L<self_gpg_keys()|https://docs.gitlab.com/ce/api/users.html#list-all-gpg-keys>

=item L<user_gpg_keys()|https://docs.gitlab.com/ce/api/users.html#single-ssh-key>

    $keys = $gitlab->self_gpg_keys();
    $keys = $gitlab->user_gpg_keys( :uid );

Returns a list of GPG keys for the authenticated user or the user with the
given C<uid>.

=item L<self_get_gpg_key()|https://docs.gitlab.com/ce/api/users.html#get-a-specific-gpg>

=item L<user_get_gpg_key()|https://docs.gitlab.com/ce/api/users.html#get-a-specific-gpg-key-for-a-given-user>

    $key = $gitlab->self_get_ssh_key( :keyid );
    $key = $gitlab->user_get_gpg_key( :uid, :keyid );

For the user that makes the request, the function returns the GPG key with the given C<keyid>.

=item L<self_add_gpg_key()|https://docs.gitlab.com/ce/api/users.html#add-a-gpg-key>

=item L<user_add_gpg_key()|https://docs.gitlab.com/ce/api/users.html#add-a-gpg-key-for-a-given-user>

    $key = $gitlab->self_add_gpg_key( :title, :key );
    $key = $gitlab->user_add_gpg_key( :uid, :title, :key );

Adds a new key for the user that makes the request or the given user.
Required arguments are

    key                 public GPG key

=item L<self_delete_gpg_key()|https://docs.gitlab.com/ce/api/users.html#delete-a-gpg-key>

=item L<user_delete_gpg_key()|https://docs.gitlab.com/ce/api/users.html#delete-a-gpg-key-for-a-given-user>

    $gitlab->self_delete_gpg_key( :keyid );
    $gitlab->user_delete_gpg_key( :uid, :keyid );

Removes the GPG key with the given C<keyid> for the user that makes the request
or the given user.

=back

=head2 E-mails

=over

=item L<self_get_emails()|https://docs.gitlab.com/ce/api/users.html#list-emails>

    $emails = $gitlab->self_get_emails();

Returns a list of B<secondary> e-mails for the user that makes the request.

Note that primary e-mail can be obtained from the result of C<whoami>.

=item L<user_get_emails()|https://docs.gitlab.com/ce/api/users.html#list-emails-for-user>

    $emails = $gitlab->user_get_emails( :uid );

Returns a list of B<secondary> e-mails for the user with the C<uid>.

Note that primary e-mail can be obtained from the result of C<user_by_id>.

=item L<self_get_email()|https://docs.gitlab.com/ce/api/users.html#single-email>

    $email = $gitlab->self_get_email( :eid );

Returns the e-mail with the given C<eid> of the user that makes the request.

=item L<self_add_email()|https://docs.gitlab.com/ce/api/users.html#add-email>

    $email = $gitlab->self_add_email( :email );

Adds a new C<email> for the user that makes the request. The e-mail must
not exist in GitLab.

=item L<user_add_email()|https://docs.gitlab.com/ce/api/users.html#add-email-for-user>

    $email = $gitlab->user_add_email( :uid, :email );

Adds a new C<email> for the user with the given C<uid>. The e-mail must
not exist in GitLab.

=item L<self_delete_email()|https://docs.gitlab.com/ce/api/users.html#delete-email-for-current-user>

    $gitlab->self_delete_email( :eid );

Deletes the e-mail with C<eid> of the user that makes the request.

=item L<user_delete_email()|https://docs.gitlab.com/ce/api/users.html#delete-email-for-given-user>

    $gitlab->user_delete_email( :uid, :eid );

Removes the e-mail with C<eid> of the user with the given C<uid>.

=back

=head2 Impersonation tokens

B<All of the following methods require administrator permissions.>

=over

=item L<user_tokens()|https://docs.gitlab.com/ce/api/users.html#get-all-impersonation-tokens-of-a-user>

    $tokens = $gitlab->user_tokens( :uid );

Retrieves every impersonation token of the user.

=item L<user_get_token()|https://docs.gitlab.com/ce/api/users.html#get-an-impersonation-token-of-a-user>

    $token = $gitlab->user_get_token( :uid, :tokenid );

Returns the requested impersonation token.

=item L<user_add_token()|https://docs.gitlab.com/ce/api/users.html#create-an-impersonation-token>

    $token = $gitlab->user_add_token( :uid, :name )

Generates a new impesonation token. This token can only by used to impersonate
the user and perfrom API calls.
I<The user will not see this token in his profile page settings.>

=item L<user_delete_token()|https://docs.gitlab.com/ce/api/users.html#revoke-an-impersonation-token>

    $gitlab->user_delete_token( :uid, :tokenid )

Revokes the given token.

=back

=head2 Account blocking

=over

=item L<user_block()|https://docs.gitlab.com/ce/api/users.html#block-user>

    $gitlab->user_block( :uid );

Blocks the user with C<uid> in the GitLab.

=item L<user_unblock()|https://docs.gitlab.com/ce/api/users.html#unblock-user>

    $gitlab->user_unblock( :uid );

Unblocks the user with C<uid> in the GitLab.
Note that LDAP blocked users cannot be unblocked by this function.
In that case the API would return C<403 Forbidden>.

=back

=head2 Miscellaneous

=over

=item L<user_events()|https://docs.gitlab.com/ce/api/users.html#get-user-contribution-events>

This method has been moved to
L<GitLab::Events::user_evens()|GitLab::Events/user_events()>.

=item L<user_activities()|https://docs.gitlab.com/ce/api/users.html#get-user-activities-admin-only>

    $stats = $gitlab->user_activities();
    $stats = $gitlab->user_activities(from => "2016-01-01");

Get the last activity date for all users, sorted from oldest to newest.
B<This method is available only to administrators.>

=back

=head1 AUTHOR

Roman Lacko <L<xlacko1@fi.muni.cz>>

=head1 SEE ALSO

=over

=item L<GitLab>

Wrapper around L<GitLab::API> and other C<GitLab::*> modules.

=back
